@@ -19,21 +19,44 @@ from googleapiclient.discovery import build
 from googleapiclient.http import MediaIoBaseDownload
 from google.oauth2.service_account import Credentials
 
-#logger = getLogger(__name__)
-
 
+# ***   CONFIGURATION   ********************************************************** #
 class config:
-    credentialjson = os.path.join(os.getcwd(), "_credentials.json")
-    dbfile = os.path.join(os.getcwd(), "_gdriveplayer.db")
-    encoding = "utf8"
-    chardet_threshold = 0.95
+    credentialjson: str = None
+    dbfile: str = None
+    encoding: str = "utf8"
+    chardet_threshold: float = 0.95
+    mplayer = "mplayer"
+    ffprobe = "ffprobe"
 
-# ***   DATABASE HELPERS ************************************************************ #
+def _set_default_config():
+    # 1. Use GDRIVEAUDIO_DIRECTORY env variable as the project root
+    # 2. If 1 is not available, use currenct working directory as the project root
+    workdir = os.environ.get("GDRIVEAUDIO_DIRECTORY", os.getcwd())
+    if os.path.isdir(workdir):
+        # use this directory as the working directory for this tool
+        config.credentialjson = os.path.join(workdir, "_credentials.json")
+        config.dbfile         = os.path.join(workdir, "_gdriveaudio.db")
+    config.encoding = "utf8"
+    config.chardet_threshold = 0.95
+
+_set_default_config()
+
+def _set_config(**kwargs):
+    for key, value in kwargs.items():
+        if not hasattr(config, key):
+            print("'%' is not a valid config name; skipped")
+            continue
+        setattr(config, key, value)
+# ***   END OF CONFIGURATION   *************************************************** #
+
+
+# ***   DATABASE HELPERS   ********************************************************** #
 AudioFile = namedtuple("AudioFile", "id name mimetype parent size md5checksum")
 AudioMeta = namedtuple("AudioMeta", "id title artist album album_artist date year genre duration")
 Folder    = namedtuple("Folder",    "id name parent fullpath")
 
-def _initialized()-> bool:
+def _database_exists()-> bool:
     return os.path.isfile(config.dbfile)
 
 def _get_sql(query: str, header: bool=False):
@@ -63,7 +86,7 @@ def _validate_sql(query: str, value=None, values=None)-> tuple:
         return True, None
     except Exception as e:
         return False, e
-# ***   END OF DATABASE HELPERS ***************************************************** #
+# ***   END OF DATABASE HELPERS   *************************************************** #
 
 
 # ***   GOOGLE DRIVE HELPERS   ****************************************************** #
@@ -240,7 +263,7 @@ def _guess_encoding(x: bytes)-> str:
     return enc
 
 def _get_audiometa(filepath: str)-> dict:
-    command = ["ffprobe", "-v", "quiet", "-print_format", "json", "-show_format", filepath]
+    command = [config.ffprobe, "-v", "quiet", "-print_format", "json", "-show_format", filepath]
     p = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
     x = p.stdout
     enc = _guess_encoding(x)
@@ -271,8 +294,24 @@ def _get_audiometa(filepath: str)-> dict:
     return out
 
 def _play_audiofile(filepath: str):
-    command = ["mplayer", "-vo", "null", filepath]
+    command = [config.mplayer, "-vo", "null", filepath]
     p = subprocess.run(command)
+
+def _check_mplayer():
+    return _check_command([config.mplayer, "--help"])
+
+def _check_ffprobe():
+    return _check_command([config.ffprobe, "-version"])
+
+def _check_command(command: list)-> bool:
+    #print(command)
+    try:
+        p = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+        #print(p)
+    except Exception as e:
+        raise ValueError("'%s' does not seem a valid command; '%s' failed with error:: %s" % (
+            config.ffprobe, " ".join(command), e))
+    return True
 # ***   END OF PLAYER HELPERS   ******************************************************* #
 
 
@@ -320,18 +359,55 @@ def init_database():
     SELECT
       a.*,
       f.name AS folder, f.fullpath || '/' AS prefix,
-      m.title, m.artist, m.album_artist, m.date, m.year, m.duration
+      m.title, m.artist, m.album, m.album_artist, m.genre, m.date, m.year, m.duration
     FROM
       audiofiles AS a
       LEFT JOIN audiometa AS m ON a.id = m.id
       LEFT JOIN folders   AS f ON a.parent = f.id
     """)
 
+def _compile_keyword(keyword, case_sensitive=False):
+    q = """SELECT name FROM pragma_table_info('audio') WHERE type LIKE 'TEXT'"""
+    textcols = [row[0] for row in _get_sql(q)]
+    textcols = [t for t in textcols if t not in ("md5checksum",)]  # disallow search on these columns
+    #print(textcols)
+    # field-specifig search
+    r = re.search(r"([^:]+):(.*)", keyword)
+    if r is not None:
+        print(r.groups())
+        field, word = r.group(1), r.group(2)
+        if field not in textcols:
+            warnings.warn("'%s' is not a valid field name thus ignored (must be one of %s)" % (field, textcols))
+        else:
+            if case_sensitive:
+                return "{} GLOB '*{}*'".format(field, word)
+            else:
+                return "{} LIKE '%{}%'".format(field, word)
+    # search over all fields
+    if case_sensitive:
+        return " OR ".join("{} GLOB '*{}*'".format(field, keyword) for field in textcols)
+    else:
+        return " OR ".join("{} LIKE '%{}%'".format(field, keyword) for field in textcols)
+
+def _compile_filter(query: str=None, keywords :list=None, keywords_case_sensitive: list=None):
+    filters = []
+    if query is not None:
+        filters.append(query)    
+    if keywords is not None:
+        for k in keywords:
+            filters.append(_compile_keyword(k, False))
+    if keywords_case_sensitive is not None:
+        for k in keywords_case_sensitive:
+            filters.append(_compile_keyword(k, True))
+    return " AND ".join("(%s)" % f for f in filters) if len(filters) > 0 else None
+
 
 def play_audio(filter: str=None, repeat: bool=False):
-    if not _initialized():
-        print("Database not initialized. Run 'gdriveaudio update -U' first")
+    _check_mplayer()
+    if not _database_exists():
+        print("Database '%s' file not found. Run 'gdriveaudio update -U' first" % config.dbfile)
         return
+
     q = "SELECT id, name, prefix FROM audio"
     if filter is not None:
         q += " WHERE {}".format(filter)
@@ -357,8 +433,17 @@ def play_audio(filter: str=None, repeat: bool=False):
             print("Finished playing all files")
             break
 
-def show_data(n: int=None, filter: str=None):
-    q = "SELECT * FROM audio"
+def show_data(n: int=None, columns: list=None, filter: str=None):
+    if not _database_exists():
+        print("Database '%s' file not found. Run 'gdriveaudio update -U' first" % config.dbfile)
+        return
+    q = """SELECT name FROM pragma_table_info('audio')"""
+    audio_cols = set([row[0] for row in _get_sql(q, header=True)])
+    for c in columns:
+        assert c in audio_cols, "'%s' is not a valid column name, must be one of %s" % (c, audio_cols)
+
+    q = "SELECT {} FROM audio".format("*" if columns is None else ",".join('"%s"' % c for c in columns))
+    print(q)
     if filter is not None:
         q += " WHERE {}".format(filter)
     if n is not None:
@@ -372,7 +457,7 @@ def show_data(n: int=None, filter: str=None):
         writer.writerow(row)
 
 def update_audio_data(files: bool=False, meta: bool=False, replace_meta: bool=False, folders: bool=False):
-    if not _initialized():
+    if not _database_exists():
         print("Initializing database")
         init_database()
     if files:
@@ -392,11 +477,15 @@ def _update_audiofiles():
     _exec_sql(q, values=tqdm(search_audio_files()))
 
 def _update_audiometa(replace: bool = False):
+    _check_ffprobe()
     if replace:
         q = "SELECT id, name FROM audiofiles"
     else:
         q = "SELECT id, name FROM audiofiles WHERE id NOT IN (SELECT id FROM audiometa)"    
     files = [(row[0], row[1]) for row in _get_sql(q)]
+    if len(files)==0:
+        print("No audio files found to update the meta data")
+        return
     ids, names = zip(*files)
     total = len(ids)
     placeholder = ",".join("?" * len(AudioMeta._fields))
@@ -423,44 +512,76 @@ def _update_folders():
 def main():
     parser = ArgumentParser(description="Play music files in google drive")
     subparsers = parser.add_subparsers(dest="command")
-    
+
     parent_parser = ArgumentParser(add_help=False)
-    parent_parser.add_argument("-c", "--credential-json", type=str, default="_credentials.json",
-                               help="Path to the google cloud credential JSON file with google drive permission")
-    parent_parser.add_argument("-d", "--database-file", type=str, default="_gdriveplayer.db",
-                               help="Path to the sqlite database file")
-    
+    parent_parser.add_argument("-D", "--workdir", type=str, default=None,
+                               help=("Directory name of the project files (_credentials.json and _gdriveaudio.db"
+                                    ". Deafult: 'GDRIVEAUDIO_DIRECTORY' environment variable, if given"
+                                    ".          Otherwise the current directory"))
+    parent_parser.add_argument("-c", "--credential-json", type=str, default=None,
+                               help="Override the path to the google cloud credential JSON file with google drive permission")
+    parent_parser.add_argument("-d", "--database-file", type=str, default=None,
+                               help="Override the path to the sqlite database file")
+
     init = subparsers.add_parser("init", parents=[parent_parser],
                                  help="Initialize database (all existing data will be deleted)")
-     
+
     update = subparsers.add_parser("update", help="Update data", parents=[parent_parser])
     update.add_argument("-U", "--update-filelist", action="store_true", help="Update file list")
     update.add_argument("-M", "--update-meta", action="store_true", help="Update audio metadata")
     update.add_argument("-F", "--update-folders", action="store_true", help="Update folder structure data")
     update.add_argument("--replace-meta", action="store_true", help="Replace existing metadata")
+    update.add_argument("--metadata-encoding", type=str, default="utf8", help="Default encoding for audio metadata")
+    update.add_argument("--chardet-threshold", type=float, default=0.95, help="Threshold to trust the chardet result")
+    update.add_argument("--ffprobe", type=str, default="ffprobe", help="ffprobe command name")
 
-    play = subparsers.add_parser("play", help="Play audio", parents=[parent_parser])
-    play.add_argument("-q", "--filter-query", type=str, default=None, help="SQL query to select files to play")
+    search = ArgumentParser(add_help=False)
+    search.add_argument("-k", "--keyword", type=str, nargs="*",
+                        help="keyword(s) to search, case-insensitive (name:word form to search in a specific field)")
+    search.add_argument("-K", "--keyword-case-sensitive", type=str, nargs="*",
+                        help="keyword(s) to search, case-sensitive (name:word form to search in a specific field)")
+    search.add_argument("-q", "--filter-query", type=str, default=None, help="SQL query to select files to show")
+
+    play = subparsers.add_parser("play", help="Play audio", parents=[search, parent_parser])
     play.add_argument("--repeat", action="store_true", help="Repeat forever")
+    play.add_argument("--mplayer", type=str, default="mplayer", help="mplayer command name")
 
-    data = subparsers.add_parser("data", help="Show data in csv format", parents=[parent_parser])
+    data = subparsers.add_parser("data", help="Show data in csv format", parents=[search, parent_parser])
     data.add_argument("-n", type=int, default=None, help="Number of rows to show")
-    data.add_argument("-q", "--filter-query", type=str, default=None, help="SQL query to select files to show")
+    data.add_argument("--columns", type=str, nargs="+", help="Columns to show")
 
     args = parser.parse_args()
     #print(args)
-    config.credentialjson = os.path.abspath(args.credential_json)
-    config.dbfile = os.path.abspath(args.database_file)
+    if args.command is None:
+        # called without subcommand, show help and finish
+        parser.print_help()
+        return
+
+    # Credentials and database file locations
+    # 1. If '--credential-json' or '--database-file' is given, use these values
+    # 2. If '--workdir' is given, then use '{workdir}/_credentials.json' and '{workdir}/_gdriveaudio.db'
+    # 3. Use '_credentials.json' and '_gdriveaudio.db' in the currenct directory
+    if args.workdir is not None:
+        _set_config(credentialjson=os.path.join(args.workdir, "_credentials.json"),
+                    dbfile=os.path.join(args.workdir, "_gdriveaudio.db"))
+    if args.credential_json is not None:
+        _set_config(credentialjson=args.credential_json)
+    if args.database_file is not None:
+        _set_config(dbfile=args.database_file)
 
     if args.command == "init":
         init_database()
     elif args.command == "update":
+        _set_config(encoding=args.metadata_encoding, chardet_threshold=args.chardet_threshold, ffprobe=args.ffprobe)
         update_audio_data(files=args.update_filelist, meta=args.update_meta,
                           replace_meta=args.replace_meta, folders=args.update_folders)
     elif args.command == "play":
-        play_audio(filter=args.filter_query, repeat=args.repeat)
+        _set_config(mplayer=args.mplayer)
+        filter = _compile_filter(query=args.filter_query, keywords=args.keyword, keywords_case_sensitive=args.keyword_case_sensitive)
+        play_audio(filter=filter, repeat=args.repeat)
     elif args.command == "data":
-        show_data(n=args.n, filter=args.filter_query)
+        filter = _compile_filter(query=args.filter_query, keywords=args.keyword, keywords_case_sensitive=args.keyword_case_sensitive)
+        show_data(n=args.n, columns=args.columns, filter=filter)
 # ***   END OF MAIN PROCEDURE   ******************************************************* #
 
 
